// @ts-ignore
import { dom, node } from 'jsx-pragmatic';
// @ts-ignore
import zoid, { destroy } from 'zoid';

export interface Theme {
  primaryColor: string;
  secondaryColor: string;
  fontFamily: string;
  logo: string;
}

export enum TransactionTypeAllowed {
  ALL = 'ALL',
  VASP_2_VASP_ONLY = 'VASP_2_VASP_ONLY',
  SELF_TRANSACTION_ONLY = 'SELF_TRANSACTION_ONLY',
}

export enum NonCustodialDeclarationType {
  SIGNATURE = 'SIGNATURE',
  DECLARATION = 'DECLARATION',
  SIGNATURE_OR_DECLARATION = 'SIGNATURE_OR_DECLARATION',
}

interface NotabeneConfig {
  widget: string;
  container: string;
  authToken: string;
  vaspDID: string;
  onValidStateChange(isValid: boolean): void;
  onError(err: any): void;
  theme?: Theme;
  dictionary?: { [key: string]: string };
  transactionTypeAllowed?: TransactionTypeAllowed;
  nonCustodialDeclarationType?: NonCustodialDeclarationType;
  beneficiaryDetails?: string[];
}

export interface OwnershipProofInput {
  type?: string | null;
  proof?: string | null;
}

export interface TransactionBlockchainInfo {
  destination?: string | null;
}
interface Transaction {
  transactionAsset: string;
  transactionAmount: string;
  beneficiaryAccountNumber: string;
  originatorVASPdid?: string;
  originatorDid?: string;
  transactionBlockchainInfo?: TransactionBlockchainInfo;
  beneficiaryVASPname?: string;
  beneficiaryName?: string;
  beneficiaryGeographicAddress?: string;
  beneficiaryNationalIdentification?: string;
  beneficiaryDateAndPlaceOfBirth?: string;
  beneficiaryDid?: string;
  isNonCustodial?: boolean;
  originatorEqualsBeneficiary?: boolean;
  beneficiaryProof?: OwnershipProofInput;
  isValid?: boolean;
  errors?: string[];
}

export default class Notabene {
  private widget: zoid.ZoidComponent;
  private popup: zoid.ZoidComponent;
  private config: NotabeneConfig;
  private tx: Transaction | undefined;
  private error?: Error;
  private widgetRendered: boolean;
  private popupRendered: boolean;

  constructor(config: NotabeneConfig) {
    this.config = config;
    this.config.nonCustodialDeclarationType =
      config.nonCustodialDeclarationType ||
      NonCustodialDeclarationType.SIGNATURE;
    this.config.transactionTypeAllowed =
      config.transactionTypeAllowed || TransactionTypeAllowed.ALL;
    this.widgetRendered = false;
    this.popupRendered = false;
  }

  public async renderWidget(): Promise<void> {
    if (this.widgetRendered) {
      console.error(
        'Notabene: Widget already rendered, please destroy it first'
      );
      return;
    }

    this.widget = await this.buildWidget();
    this.popup = await this.buildPopup();
    this.widget.updateProps({
      transaction: this.tx,
    });
    this.popup.updateProps({
      transaction: this.tx,
    });
    this.popup.hide();
    await this.widget.render(this.config.container);
    await this.renderPopup();
  }

  private async renderPopup(): Promise<void> {
    if (!this.popup) return;

    if (!this.popupRendered) {
      await this.popup.render(this.config.container);
      this.popupRendered = true;
    }
  }

  public async destroyWidget(): Promise<void> {
    await destroy();
    this.widgetRendered = false;
    this.popupRendered = false;
  }

  public setTransaction(tx: Transaction) {
    const validProperties = [
      'transactionAsset',
      'transactionAmount',
      'beneficiaryAccountNumber',
      'beneficiaryAccountNumber',
      'originatorVASPdid',
      'originatorDid',
      'transactionBlockchainInfo',
      'beneficiaryVASPname',
      'beneficiaryVASPdid',
      'beneficiaryName',
      'beneficiaryGeographicAddress',
      'beneficiaryNationalIdentification',
      'beneficiaryDateAndPlaceOfBirth',
      'beneficiaryDid',
      'isNonCustodial',
      'originatorEqualsBeneficiary',
      'beneficiaryProof',
      'isValid',
      'errors',
    ];
    if (tx) {
      Object.keys(tx).forEach((key) => {
        if (!validProperties.includes(key)) {
          throw new Error(`Notabene: Invalid transaction property ${key}`);
        }
      });
    }
    this.tx = tx;
    if (this.widgetRendered) {
      this.widget.updateProps({
        transaction: tx,
      });
      this.popup.updateProps({
        transaction: tx,
      });
    }
  }

  private getWidgetUrl(): string {
    const url = new URL(this.config.widget);
    url.hash = `authToken=${this.config.authToken}`;
    return url.toString();
  }

  private getAdditionalInfoUrl(): string {
    const url = new URL(this.config.widget);
    url.search = 'popup=true';
    url.hash = `authToken=${this.config.authToken}`;
    return url.toString();
  }

  private handleError(err: any): void {
    this.error = err;
    if (this.widget && this.popup) {
      this.widget.updateProps({
        error: err,
      });
      this.popup.updateProps({
        error: err,
      });
    }
    if (!err || !this.config.onError) return;
    this.config.onError(err);
  }

  private buildWidget(): zoid.ZoidComponent {
    if (!this.config.authToken) {
      console.error('Notabene: authToken is required');
      return;
    }

    const builtWidget = zoid.create({
      tag: 'widget',
      url: this.getWidgetUrl(),
      dimensions: {
        width: '100%',
        height: '280px',
      },
      props: {
        vaspDID: {
          type: 'string',
          required: true,
        },
        transaction: {
          type: 'object',
          required: false,
        },
        onRequireAdditionalInformation: {
          type: 'function',
          required: true,
        },
        onValidStateChange: {
          type: 'function',
          required: true,
        },
        error: {
          type: 'object',
          required: false,
        },
        onError: {
          type: 'function',
          required: true,
        },
        storeTransaction: {
          type: 'function',
          required: true,
        },
        theme: {
          type: 'object',
          required: false,
        },
        dictionary: {
          type: 'object',
          required: false,
        },
        transactionTypeAllowed: {
          type: 'string',
          required: false,
        },
        nonCustodialDeclarationType: {
          type: 'string',
          required: false,
        },
        beneficiaryDetails: {
          type: 'array',
          required: false,
        },
      },
    })({
      vaspDID: this.config.vaspDID,
      transaction: {},
      theme: this.config.theme,
      dictionary: this.config.dictionary,
      transactionTypeAllowed: this.config.transactionTypeAllowed,
      nonCustodialDeclarationType: this.config.nonCustodialDeclarationType,
      beneficiaryDetails: this.config.beneficiaryDetails,
      onRequireAdditionalInformation: () => {
        this.popup.updateProps({
          popupOpen: true,
        });
        this.popup.show();
      },
      onValidStateChange: (isValid: boolean) => {
        this.config.onValidStateChange(isValid);
      },
      error: this.error,
      onError: (err: any) => {
        this.handleError(err);
        console.debug('Notabene: Error handler called');
      },
      storeTransaction: (tx: Transaction) => {
        this.setTransaction(tx);
      },
    });

    builtWidget.event.on(zoid.EVENT.RENDERED, (err: any) => {
      this.widgetRendered = true;
    });

    return builtWidget;
  }

  private buildPopup(): zoid.ZoidComponent {
    if (!this.config.authToken) {
      console.error('Notabene: authToken is required');
      return;
    }

    const builtPopup = zoid.create({
      tag: 'popup',
      url: this.getAdditionalInfoUrl(),
      children: () => {
        return {
          Transaction: this.tx,
        };
      },
      dimensions: {
        width: '100%',
        height: '100%',
      },
      props: {
        vaspDID: {
          type: 'string',
          required: true,
        },
        transaction: {
          type: 'object',
          required: false,
        },
        storeTransaction: {
          type: 'function',
          required: true,
        },
        close: {
          type: 'function',
        },
        theme: {
          type: 'object',
          required: false,
        },
        dictionary: {
          type: 'object',
          required: false,
        },
        transactionTypeAllowed: {
          type: 'string',
          required: false,
        },
        nonCustodialDeclarationType: {
          type: 'string',
          required: false,
        },
        popupOpen: {
          type: 'boolean',
          required: true,
        },
        setPopupOpen: {
          type: 'function',
          required: true,
        },
        error: {
          type: 'object',
          required: false,
        },
        onError: {
          type: 'function',
          required: true,
        },
        beneficiaryDetails: {
          type: 'array',
          required: false,
        },
      },
      containerTemplate: ({ uid, doc, frame, prerenderFrame }: any) => {
        return node(
          'div',
          { id: uid, class: 'nb-container' },
          node(
            'style',
            null,
            `
            #${uid}.container {
                border: 5px solid red;
            }
            .nb-overlay {
              position: fixed;
              width: 100vw;
              height: 100vh;
              top: 0;
              left: 0;
              display: flex;
              align-items: center;
              justify-content: center;
              background-color: rgba(0,0,0,0.5);
            }
            .nb-popup {
              max-width: 608px;
              width: 100%;
              height: 100%;
            }
            .nb-preloader {
              display: none;
            }
          `
          ),
          node(
            'div',
            { class: 'nb-overlay' },
            node('node', { el: frame, class: 'nb-popup' }),
            node('node', { el: prerenderFrame, class: 'nb-preloader' })
          )
        ).render(dom({ doc }));
      },
    })({
      vaspDID: this.config.vaspDID,
      transaction: this.tx,
      beneficiaryDetails: this.config.beneficiaryDetails,
      storeTransaction: (tx: Transaction) => {
        this.setTransaction(tx);
        console.debug('Notabene: Transaction stored');
      },
      close: () => {
        this.popup.close();
        console.debug('Notabene: Popup closing');
      },
      setPopupOpen: (isOpen: boolean) => {
        this.popup.updateProps({
          popupOpen: isOpen,
        });
      },
      error: this.error,
      onError: (err: any) => {
        this.handleError(err);
        console.debug('Notabene: Error handler called');
      },
      popupOpen: false,
      theme: this.config.theme,
      dictionary: this.config.dictionary,
      transactionTypeAllowed: this.config.transactionTypeAllowed,
      nonCustodialDeclarationType: this.config.nonCustodialDeclarationType,
    });

    builtPopup.event.on(zoid.EVENT.RENDERED, (err: any) => {
      this.popupRendered = true;
    });

    return builtPopup;
  }
}

if (window) window.Notabene = Notabene;
